function [x, y, out] = pdNCG_ls(opts)
%PDNCG pdNCG solver for the l1-regularized problem
%
% min tau*||x||_1 + 0.5||Ax-b||_2^2,
%
% where matrix A^T*A is fullrank. 

%% Initialize output.
out.itr  = 1; % pdNCG iterations.
out.f = sparse(opts.maxpdNCGitr,1); % Objective values at every iteration.
out.t = sparse(opts.maxpdNCGitr,1); % time passed at every iteration.

%% Check if A'*A = I, i.e. denoising.
tmp = rand(opts.n,1);
if norm(opts.A'*(opts.A*tmp)-tmp,1)/norm(tmp,1) < 1e-15;
    opts.reg = 1.0e-16;
end
clear tmp;

%% Initialize data.
stagnation = 0;

% PCG data structure that carries everything the call to the
% preconditioned CG method needs.
PCG = [];

% Initialize diagonal preconditioner if needed.
if strcmp(opts.PCG_method,'ls_diag_pcg')
    if isnumeric(opts.A)
        PCG.diag_ATA = sum(opts.A.^2,1)';
    else
        PCG.diag_ATA = 0.5.*ones(opts.n,1);
    end
end

nwtn = zeros(opts.n,1);

x = opts.x0;

D  = 1./sqrt(opts.mu^2 + x.^2);
Dx = D.*x;

y = opts.y0;

% Calculate the approximate Hessian of pseudo-Huber function.
H  = D.*(1 - Dx.*y);
        
tauH = opts.tau.*H;

% Initialize diagonal preconditioner if needed.
if strcmp(opts.PCG_method,'ls_diag_pcg')
    PCG.tauH = tauH;
end

% Calculate gradient at x0.
g_ls = opts.A.'*(opts.A*x-opts.b);

g = opts.tau.*Dx + g_ls;

% Fidelity and regularization functions.
f_ls    = @(q) 0.5*norm(opts.A*q-opts.b)^2;
f_huber = @(q) opts.tau*sum(sqrt(opts.mu^2 + q.^2) - opts.mu);
    
%% Print
if opts.verbose >= 1
    fprintf('%5s | %5s    | %5s    | %5s    | %5s | %5s |  %5s  \n',...
        'pdNCG itr', 'a', 'tau', 'mu', '|d|_x/|x|','PCG itr', 'PCG rs');
end

%% Start Newton conjugate gradients.
while 1        
    % Define regularized second-order derivative.
    Hessian = @(q) tauH.*q + opts.A'*(opts.A*q) + opts.reg*q;
    
    % Call user function.
    if isfield(opts, {'user_fcn'}) 
        opts.user_fcn(opts,x,y,@(q) f_huber(q) + f_ls(q),g,Hessian,nwtn);
    end  
    
    % Calculate Newton direction.  
	[nwtn, rlrs, pcgi] = PCG_wrapper(PCG, Hessian, g, opts);      
    
    normN_scaled = -(nwtn'*g + opts.reg*norm(nwtn)^2)/norm(x);    
        
    dy = H.*nwtn - (y - Dx);
        
    y_tilde = y + dy;
    y = sign(y_tilde).*min(abs(y_tilde),1);
        
    % Calculate stepsize and update variables.        
    [x, a, f_new] = simple_bt(x, nwtn, f_ls, f_huber, normN_scaled, opts);
    
    % Store objective function and corresponding time.
    out.f(out.itr) = f_new;
    out.t(out.itr) = toc; 
    
    % The PCG methods needs to know the stepsize.
    PCG.a = a;
        
    % Printing info.
    if opts.verbose >= 1
        fprintf('  %4d    | %1.2e | %1.2e | %1.2e |  %8s | %4d    | %1.2e \n',...
            out.itr, a, opts.tau, opts.mu, num2str(normN_scaled,'%1.2e\n'), pcgi, rlrs);
    end
        
    % Calculate hessian of pseudo-Huber function.
    D  = 1./sqrt(opts.mu^2 + x.^2);
    Dx = D.*x;
    H  = D.*(1 - Dx.*y);
        
    tauH = opts.tau.*H;
    
    % Initialize diagonal preconditioner if needed.
    if strcmp(opts.PCG_method,'ls_diag_pcg')
        PCG.tauH = tauH;
    end
        
    % Calculate first-order derivative of perturbed problem.
    g_ls = opts.A.'*(opts.A*x-opts.b);

    g     = opts.tau.*Dx + g_ls;
    
    if isfield(opts, {'user_stop'}) 
        if opts.user_stop(opts,x,y,@(q) f_huber(q) + f_ls(q),g,...
                          Hessian,nwtn)
            out.message = 'Converged';
            break;
        end
    else
        if normN_scaled <= opts.tol
            out.message = 'Converged';
            break; 
        end         
    end
    
    % Maximum iterations reached.
    if out.itr > opts.maxpdNCGitr
        stagnation = 1;
    end
    
    if stagnation == 1
       out.message = 'Maximum iterations of pdNCG reached';
       fprintf('Maximum iterations of pdNCG reached, termination.\n'); 
       fprintf('Possible fixes: \n');
       fprintf('a) increase opts.maxpdNCGitr. Performs more pdNCG iterations.\n');
       fprintf('b) decrease opts.tolpcg. Calculates better directions. \n');
       fprintf('c) increase opts.mxitrpcg. \n');
       break;
    end 
    
    % Update iteration counter.
    out.itr = out.itr + 1;
end
%% End of primal-dual Newton conjugate gradients.

out.f = out.f(1:out.itr);
out.t = out.t(1:out.itr);

% Call user function.
if isfield(opts, {'user_fcn'}) 
    opts.user_fcn(opts,x,y,@(q) f_huber(q) + f_ls(q),g,Hessian,nwtn);
end 